#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <pthread.h>

#include "rocksdb/db.h"
#include "rocksdb/utilities/transaction_db.h"

rocksdb::TransactionDB *rdb;

std::vector<rocksdb::ColumnFamilyDescriptor> cf_descr;
std::vector<rocksdb::ColumnFamilyHandle *> cf_handles;

rocksdb::ColumnFamilyHandle * default_cf_handle;

int N_TRANSACTIONS=1000;
int N_THREADS=1;

int SYNC_COMMIT= true;
int SYNC_PREPARE= true;
int USING_2PC= true;

/*
  Transaction function. Do N RocksDB transactions.
*/
void *transaction_func(void *thread_id)
{
  rocksdb::WriteOptions write_opts;
  write_opts.sync= SYNC_COMMIT;
  write_opts.disableWAL= false;
  bool error= false;
  
  for (int trx_no=0 ; trx_no < N_TRANSACTIONS; trx_no++)
  {
    rocksdb::Transaction *rocksdb_tx;
    rocksdb_tx = rdb->BeginTransaction(write_opts);

    // Make a few write operations
    for (int i = 0; i < 10; i++) {
      int key_buf;
      int val_buf;
      key_buf= htonl(rand());
      val_buf= htonl(rand());

      const rocksdb::Status s= 
        rocksdb_tx->Put(default_cf_handle,
                        rocksdb::Slice((char*)&key_buf, sizeof(int)),
                        rocksdb::Slice((char*)&val_buf, sizeof(int)));
      if (!s.ok()) {
        fprintf(stderr, "Put error %s\n", s.ToString().c_str());
        error= true;
        break;
      }
    }

    if (error)
      break;

    // Optional: prepare first
    if (USING_2PC)
    {
      int name_buf;
      name_buf= htonl(rand());
      rocksdb::Status s;
      s = rocksdb_tx->SetName(rocksdb::TransactionName((char*)&name_buf, sizeof(int)));
      if (!s.ok()) {
        fprintf(stderr, "SetName error %s\n", s.ToString().c_str());
        break;
      }

      rocksdb::WriteBatch* wb= rocksdb_tx->GetCommitTimeWriteBatch();
      int key_buf;
      int val_buf;
      key_buf= htonl(rand());
      val_buf= htonl(rand());

      s= wb->Put(default_cf_handle,
                 rocksdb::Slice((char*)&key_buf, sizeof(int)),
                 rocksdb::Slice((char*)&val_buf, sizeof(int)));
      if (!s.ok()) {
        fprintf(stderr, "Put error %s\n", s.ToString().c_str());
        error= true;
        break;
      }

      rocksdb::WriteOptions write_opts;
      write_opts.sync= SYNC_PREPARE;
      write_opts.disableWAL= false;
      rocksdb_tx->SetWriteOptions(write_opts);

      s = rocksdb_tx->Prepare();
      if (!s.ok()) {
        fprintf(stderr, "Prepare error %s\n", s.ToString().c_str());
        break;
      }

    }

    // Commit
    {
      //void SetWriteOptions(const WriteOptions& write_options) = 0;
      rocksdb::WriteOptions write_opts;
      write_opts.sync= SYNC_COMMIT;
      write_opts.disableWAL= false;
      rocksdb_tx->SetWriteOptions(write_opts);

      const rocksdb::Status s = rocksdb_tx->Commit();
      if (!s.ok()) { fprintf(stderr, "Commit error %s\n", s.ToString().c_str());
        break;
      }
    }
  }

  pthread_exit(NULL);
}

int main(int argc, char **argv)
{
  if (argc != 6)
  {
    fprintf(stderr, 
             "Usage: test-gcommit N_TRANSACTIONS N_THREADS USING_2PC SYNC_PREPARE SYNC_COMMIT\n");
    exit(1);
  }
  sscanf(argv[1], "%d", &N_TRANSACTIONS);
  sscanf(argv[2], "%d", &N_THREADS);
  sscanf(argv[3], "%d", &USING_2PC);
  sscanf(argv[4], "%d", &SYNC_PREPARE);
  sscanf(argv[5], "%d", &SYNC_COMMIT);

  rocksdb::Options main_opts;
  main_opts.create_if_missing= true;

  rocksdb::TransactionDBOptions tx_db_options;
  const char *rocksdb_datadir="/home/psergey/rocksdb-sept-data1";
  rocksdb::Status status;
  std::vector<std::string> cf_names;
  
  status = rocksdb::DB::ListColumnFamilies(main_opts, rocksdb_datadir,
                                           &cf_names);

  for (size_t i = 0; i < cf_names.size(); ++i) {
    rocksdb::ColumnFamilyOptions opts;
    cf_descr.push_back(rocksdb::ColumnFamilyDescriptor(cf_names[i], opts));
  }

  status = rocksdb::TransactionDB::Open(main_opts, tx_db_options, 
                                        rocksdb_datadir, 
                                        cf_descr, &cf_handles,
                                        &rdb);
  default_cf_handle= cf_handles[0];
  if (!status.ok()) {
    fprintf(stderr, "Error opening database\n");
    exit(1);
  }

  fprintf(stderr, "# RocksDB Instance Opened\n");

  fprintf(stderr, "Starting run using %d threads doing %d transactions each\n",
          N_THREADS, N_TRANSACTIONS);
  fprintf(stderr, "  sync_prepare=%s\n", SYNC_PREPARE? "true" :"false"); 
  fprintf(stderr, "  sync_commit=%s\n", SYNC_COMMIT? "true" :"false"); 
  fprintf(stderr, "  using_2pc=%s\n", USING_2PC? "true" :"false"); 

  pthread_t threads[2000];

  struct timespec time_start, time_end;
  clock_gettime(CLOCK_REALTIME, &time_start);
  uint i;

  for (i=0; i < N_THREADS; i++)
  {
    if (pthread_create(&threads[i], NULL, transaction_func, (void *)NULL))
    {
      fprintf(stderr, "Error in pthread_join\n");
      exit(1);
    }
  }

  for (i=0; i < N_THREADS; i++)
  {
     if (pthread_join(threads[i], NULL))
       fprintf(stderr, "Error in pthread_join\n");
  }

  //transaction_func();
  
  clock_gettime(CLOCK_REALTIME, &time_end);

  const double BILLION=1e9;
  double accum = (time_end.tv_sec - time_start.tv_sec) + ( time_end.tv_nsec - time_start.tv_nsec ) / BILLION;
  printf("Took %lf seconds\n", accum);

  printf("TH\ttrx\tn_threads\ttime\ttps\n");
  printf("TR\t%d\t%d\t%lf\t%lf\n", N_TRANSACTIONS, N_THREADS, accum,
          (double(N_TRANSACTIONS)*N_THREADS/accum));

  return 0;
}

